/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file recv_multipart_list_parts.c
 */
#include "globus_gridftp_server_s3_base.h"

static
S3Status
globus_l_s3_recv_multipart_list_parts_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data);

static
void
globus_l_s3_recv_multipart_list_parts_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_list_parts_response(
    int                                 is_truncated,
    const char                         *next_part_number_marker,
    const char                         *initiator_id,
    const char                         *initiator_display_name,
    const char                         *owner_id,
    const char                         *owner_display_name,
    const char                         *storage_class,
    int                                 parts_count,
    int                                 last_part_number,
    const S3ListPart                   *parts,
    void                               *callback_data);

globus_result_t
globus_i_s3_recv_multipart_list_parts(
    globus_i_s3_transfer_t             *transfer_state)
{
    size_t                              first_restarted_part;

    GlobusS3Enter();

    do
    {
        char                           *next_marker;

        next_marker = transfer_state->next_marker;
        transfer_state->next_marker = NULL;

        transfer_state->retries = 0;

        do
        {
            transfer_state->status = S3StatusOK;
            transfer_state->result = GLOBUS_SUCCESS;

            globus_i_s3_retry_delay(transfer_state->retries);

            GlobusS3Info("REST: op=%p S3_list_parts(id=%s) try=%d\n",
                        (void *) transfer_state->op,
                        transfer_state->multipart_id,
                        transfer_state->retries);

            S3_list_parts(
                    &transfer_state->bucket_context,
                    transfer_state->key,
                    next_marker,
                    transfer_state->multipart_id,
                    NULL,
                    0,
                    NULL,
                    &(S3ListPartsHandler)
                    {
                        .responseHandler.propertiesCallback
                                      = globus_l_s3_recv_multipart_list_parts_properties,
                        .responseHandler.completeCallback
                                      = globus_l_s3_recv_multipart_list_parts_complete,
                        .responseXmlCallback
                                      = globus_l_s3_recv_multipart_list_parts_response
                    },
                    transfer_state);

            GlobusS3DebugOpStatus(S3_list_parts, transfer_state);
        }
        while (S3_status_is_retryable(transfer_state->status)
                && (transfer_state->retries++) < globus_i_s3_max_retries);

        free(next_marker);
    }
    while (transfer_state->result == GLOBUS_SUCCESS && transfer_state->next_marker != NULL);

    // However, the client might not have gotten all of the restart
    // markers, so we'll zero out the etags of the ones that the client is
    // going to resend us
    first_restarted_part = transfer_state->offset / transfer_state->part_size;
    for (size_t i = first_restarted_part; i < transfer_state->num_parts; i++)
    {
        transfer_state->parts[i].etag[0] = 0;
    }

    for (size_t i = 0; i < first_restarted_part; i++)
    {
        if (transfer_state->result == GLOBUS_SUCCESS
            && transfer_state->parts[i].etag[0] == 0)
        {
            transfer_state->result = GlobusGFSErrorGeneric("Missing part etag");
            break;
        }
    }

    GlobusS3ExitResult(transfer_state->result);

    return transfer_state->result;
}
/* globus_i_s3_recv_multipart_list_parts() */

static
S3Status
globus_l_s3_recv_multipart_list_parts_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_list_parts_properties() */

static
void
globus_l_s3_recv_multipart_list_parts_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    assert(transfer_state != NULL);

    if (status != S3StatusOK && transfer_state->result == GLOBUS_SUCCESS)
    {
        transfer_state->status = status;
        transfer_state->result = GlobusGFSErrorS3(status, transfer_state->transfer_info->pathname, error_details);
    }

    GlobusS3ExitResult(transfer_state->result);

    return;
}
/* globus_l_s3_recv_multipart_list_parts_complete() */

static
S3Status
globus_l_s3_recv_multipart_list_parts_response(
    int                                 is_truncated,
    const char                         *next_part_number_marker,
    const char                         *initiator_id,
    const char                         *initiator_display_name,
    const char                         *owner_id,
    const char                         *owner_display_name,
    const char                         *storage_class,
    int                                 parts_count,
    int                                 last_part_number,
    const S3ListPart                   *parts,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    assert(transfer_state != NULL);

    for (size_t i = 0; i < (size_t) parts_count; i++)
    {
        if ((parts[i].partNumber-1) > transfer_state->num_parts)
        {
            status = S3StatusInternalError;
            transfer_state->result = GlobusGFSErrorGeneric(
                    "Multipart transfer size changed");
            goto fail;
        }
        if (transfer_state->parts[parts[i].partNumber-1].etag[0] != 0)
        {
            status = S3StatusInternalError;
            transfer_state->result = GlobusGFSErrorGeneric(
                    "Duplicate parts in multipart transfer");
            goto fail;
        }
        if (transfer_state->parts[parts[i].partNumber-1].part_length != parts[i].size)
        {
            status = S3StatusInternalError;
            transfer_state->result = GlobusGFSErrorGeneric(
                    "Incorrect part size");
            goto fail;
        }
        assert(strlen(parts[i].eTag) < sizeof(transfer_state->parts[parts[i].partNumber-1].etag));
        strcpy(transfer_state->parts[parts[i].partNumber-1].etag, parts[i].eTag);
    }

    if (is_truncated && next_part_number_marker != NULL)
    {
        transfer_state->next_marker = strdup(next_part_number_marker);

        if (transfer_state->next_marker == NULL)
        {
            status = S3StatusOutOfMemory;
            transfer_state->result = GlobusGFSErrorMemory("next_marker");
            goto fail;
        }
    }

fail:
    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_list_parts_response() */
// vim: ft=c:
