/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_gridftp_server_s3_base.h"

/**
 * @file recv_single.c
 * Implementation of the S3 DSI Receive a small file in a single transfer
 */

static
void *
globus_l_s3_recv_single_thread(
    void                               *user_arg);

static
S3Status
globus_l_s3_recv_single_properties(
    const S3ResponseProperties         *properties,
    void                               *callbackData);

static
void
globus_l_s3_recv_single_complete(
    S3Status                            status,
    const S3ErrorDetails               *errorDetails,
    void                               *callbackData);

static
int
globus_l_s3_recv_single_data(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_data);

static
globus_result_t
globus_l_s3_recv_register_reads(
    globus_i_s3_transfer_t             *transfer_state);

static
void
globus_l_s3_recv_single_read_callback(
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t                      *buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void                               *user_arg);

globus_result_t
globus_i_s3_recv_single(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_i_s3_transfer_t             *transfer_state = NULL;
    globus_thread_t                     t;
    int                                 rc;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    assert(s3_handle != NULL);
    assert(transfer_info != NULL);

    GlobusS3Info("recv: op=%p pathname=%s\n", 
            (void *) op,
            transfer_info->pathname);

    transfer_state = malloc(sizeof(globus_i_s3_transfer_t) + strlen(transfer_info->pathname) + 1);
    if (transfer_state == NULL)
    {
        result = GlobusGFSErrorMemory("transfer_state");
        goto fail;
    }

    *transfer_state = (globus_i_s3_transfer_t)
    {
        .op                           = op,
        .transfer_info                = transfer_info,
        .pending_buffers_last         = &transfer_state->pending_buffers
    };
    globus_mutex_init(&transfer_state->mutex, NULL);
    globus_cond_init(&transfer_state->cond, NULL);

    strcpy(transfer_state->pathname, transfer_info->pathname);
    globus_i_s3_path_parse(transfer_state->pathname, &transfer_state->bucket_name, &transfer_state->key);

    transfer_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, transfer_state->bucket_name);

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_recv_single_thread,
            transfer_state);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");

        globus_cond_destroy(&transfer_state->cond);
        globus_mutex_destroy(&transfer_state->mutex);

fail:
        GlobusS3Info("finished_transfer: op=%p result=%#x\n",
                    (void *) op,
                    (unsigned int) result);
        
        globus_gridftp_server_finished_transfer(op, result);
        free(transfer_state);
    }

    return result;
}
/* globus_i_s3_recv_single() */

static
void *
globus_l_s3_recv_single_thread(
    void                               *user_arg)
{
    globus_i_s3_transfer_t             *transfer_state = user_arg;
    char                                s3_md5[((MD5_DIGEST_LENGTH+2)*4)/3+1];
    S3PutProperties                     put_properties;
    time_t                              modification_time;
    char                                modification_time_string[snprintf(NULL, 0, "%"PRIdMAX, (intmax_t) INTMAX_MAX)+2];
    S3NameValue                         meta_data[2];
    globus_result_t                     result;

    GlobusS3Enter();

    modification_time_string[0] = 0;
    s3_md5[0] = 0;

    globus_mutex_lock(&transfer_state->mutex);

    put_properties = (S3PutProperties)
    {
        .contentType                  = globus_i_s3_mime_type_get(transfer_state->key),
        .md5                          = s3_md5[0] ? s3_md5 : NULL,
        .metaData                     = meta_data,
        .metaDataCount                = 0
    };

    result = globus_gridftp_server_get_recv_modification_time(
            transfer_state->op,
            &modification_time);
    if (result == GLOBUS_SUCCESS && modification_time != (time_t) -1)
    {
        snprintf(modification_time_string,
                 sizeof(modification_time_string),
                 "%"PRIdMAX,
                (intmax_t) modification_time);

        meta_data[put_properties.metaDataCount++] = (S3NameValue)
        {
            .name                 = GLOBUS_S3_MTIME_NAME,
            .value                = modification_time_string
        };
    }
    if (transfer_state->transfer_info->expected_checksum_alg != NULL
        && strcasecmp(transfer_state->transfer_info->expected_checksum_alg, "md5") == 0)
    {
        if (strlen(transfer_state->transfer_info->expected_checksum) != 32)
        {
            result = GlobusGFSErrorGeneric("Invalid md5 checksum");
            goto bad_checksum;
        }

        result = globus_i_s3_md5sum_encode(
                transfer_state->transfer_info->expected_checksum, s3_md5);
        if (result != GLOBUS_SUCCESS)
        {
            goto bad_checksum;
        }
        meta_data[put_properties.metaDataCount++] = (S3NameValue)
        {
            .name                 = GLOBUS_S3_MD5_NAME,
            .value                = s3_md5
        };
    }

    GlobusS3Info("begin_transfer: op=%p pathname=%s\n",
                 (void *) transfer_state->op, transfer_state->transfer_info->pathname);
    globus_gridftp_server_begin_transfer(transfer_state->op, 0, NULL);

    result = globus_l_s3_recv_register_reads(transfer_state);
    if (result != GLOBUS_SUCCESS)
    {
        goto cant_register;
    }

    do
    {
        transfer_state->status = S3StatusOK;
        transfer_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(transfer_state->retries);

        GlobusS3Info("REST: op=%p S3_put_object(bucket=%s, key=%s) try=%d\n",
                    (void *) transfer_state->op,
                    transfer_state->bucket_name,
                    transfer_state->key,
                    transfer_state->retries);

        S3_put_object(
                &transfer_state->bucket_context,
                transfer_state->key,
                (uint64_t) transfer_state->transfer_info->alloc_size,
                &put_properties,
                NULL,
                &(S3PutObjectHandler)
                {
                    .responseHandler      =
                    {
                        .propertiesCallback
                                          = globus_l_s3_recv_single_properties,
                        .completeCallback = globus_l_s3_recv_single_complete,
                    },
                    .putObjectDataCallback
                                          = globus_l_s3_recv_single_data
                },
                transfer_state);

        GlobusS3DebugOpStatus(S3_put_object, transfer_state);
    } while (S3_status_is_retryable(transfer_state->status)
            && (transfer_state->retries++) < globus_i_s3_max_retries);

    /* Possible that libs3 called finished before we registered a 
     * callback that resulted in EOF.
     */
    if (transfer_state->registered_buffers == NULL && ! transfer_state->eof)
    {
        globus_l_s3_recv_register_reads(transfer_state);
    }
    while (transfer_state->registered_buffers != NULL)
    {
        globus_cond_wait(&transfer_state->cond,
                         &transfer_state->mutex);
    }

cant_register:
    result = transfer_state->result;

bad_checksum:
    GlobusS3Info("finished_transfer: op=%p result=%#x\n",
                 (void *) transfer_state->op,
                 (unsigned int) result);
    globus_gridftp_server_finished_transfer(transfer_state->op, result);

    globus_mutex_unlock(&transfer_state->mutex);
    globus_mutex_destroy(&transfer_state->mutex);
    globus_cond_destroy(&transfer_state->cond);

    free(transfer_state);

    return NULL;
}
/* globus_l_s3_recv_single_thread() */

static
S3Status
globus_l_s3_recv_single_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();
    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        status = S3StatusInternalError;
    }

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_single_properties() */

static
void
globus_l_s3_recv_single_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    assert(transfer_state != NULL);

    if (status != S3StatusOK && transfer_state->result == GLOBUS_SUCCESS)
    {
        transfer_state->result = GlobusGFSErrorS3(status, transfer_state->transfer_info->pathname, error_details);
        transfer_state->status = status;
    }

    GlobusS3Exit();

    return;
}
/* globus_l_s3_recv_single_complete() */

static
globus_result_t
globus_l_s3_recv_register_reads(
    globus_i_s3_transfer_t             *transfer_state)
{
    globus_i_s3_buffer_t               *buffer;
    globus_size_t                       optimal_blocksize;
    int                                 optimal_concurrency;
    int                                 currently_registered;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    buffer = transfer_state->registered_buffers;
    currently_registered = 0;

    while (buffer != NULL)
    {
        currently_registered++;
        buffer = buffer->next;
    }
    GlobusS3Trace("Currently %d blocks registered\n", currently_registered);

    if (transfer_state->result == GLOBUS_SUCCESS && !transfer_state->eof)
    {
        /* Reads still might be useful */
        globus_gridftp_server_get_block_size(
                transfer_state->op,
                &optimal_blocksize);

        globus_gridftp_server_get_optimal_concurrency(
                transfer_state->op,
                &optimal_concurrency);
        
        GlobusS3Trace("Optimal is %d blocks\n", optimal_concurrency);

        for (int i = currently_registered; i < optimal_concurrency; i++)
        {
            buffer = globus_i_s3_buffer_get(
                    transfer_state, (size_t) optimal_blocksize);
            transfer_state->current_buffer = NULL;

            if (buffer == NULL)
            {
                result = GlobusGFSErrorMemory("buffer");

                goto buffer_fail;
            }
            buffer->transfer_offset = (uint64_t) -1;

            assert(buffer->buffer_used == 0);

            GlobusS3Trace("Registering read with s3_buffer=%p\n",
                          (void *) buffer);
            result = globus_gridftp_server_register_read(
                    transfer_state->op,
                    buffer->buffer,
                    buffer->buffer_len,
                    globus_l_s3_recv_single_read_callback,
                    transfer_state);
            if (result != GLOBUS_SUCCESS)
            {
                goto register_fail;
            }
            buffer->next = transfer_state->registered_buffers;
            transfer_state->registered_buffers = buffer;
        }
    }

register_fail:
buffer_fail:
    GlobusS3Exit();
    return result;
}
/* globus_l_s3_recv_register_reads() */

/**
 * Data callback called by libs3 to request more data to put to the
 * object being stored.
 */
static
int
globus_l_s3_recv_single_data(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;
    int                                 buffer_filled = 0;
    globus_i_s3_buffer_t               *s3_buffer;

    GlobusS3Enter();

    GlobusS3Trace("recv_single_data: op=%p buffer_size=%d buffer=%p\n",
            (void *) transfer_state->op,
            buffer_size,
            (void *) buffer);

    while ((!globus_i_s3_recv_is_transfer_offset_ready(transfer_state))
           && (!globus_i_s3_recv_is_reading_complete(transfer_state)))
    {
        globus_l_s3_recv_register_reads(transfer_state);

        GlobusS3Trace("Waiting for current offset %"PRIuMAX" to be ready\n",
                      (uintmax_t) transfer_state->offset);

        globus_cond_wait(&transfer_state->cond, &transfer_state->mutex);
    }

    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        buffer_filled = -1;

        goto done;
    }

    while ((buffer_filled < buffer_size)
            && globus_i_s3_recv_is_transfer_offset_ready(transfer_state))
    {
        int                             to_copy;

        s3_buffer = transfer_state->pending_buffers;

        to_copy = buffer_size - buffer_filled;

        if (to_copy > s3_buffer->buffer_used)
        {
            to_copy = s3_buffer->buffer_used;
        }
        assert (to_copy > 0);

        memcpy(buffer+buffer_filled, s3_buffer->buffer, to_copy);
        buffer_filled += to_copy;
        transfer_state->offset += to_copy;

        if (to_copy < s3_buffer->buffer_used)
        {
            /* Partial buffer copy */
            GlobusS3Trace("partial_buffer_copy: op=%p bytes_copied=%d\n",
                    (void *) transfer_state->op,
                    to_copy);
            memmove(s3_buffer->buffer, s3_buffer->buffer + to_copy, s3_buffer->buffer_used - to_copy);
            s3_buffer->buffer_used -= to_copy;
            s3_buffer->transfer_offset += to_copy;
        }
        else
        {
            GlobusS3Trace("add_to_pending_buffers: op=%p s3_buffer=%p\n",
                    (void *) transfer_state->op,
                    (void *) s3_buffer);

            transfer_state->pending_buffers = s3_buffer->next;
            if (transfer_state->pending_buffers == NULL)
            {
                transfer_state->pending_buffers_last = &transfer_state->pending_buffers;
            }
            globus_i_s3_buffer_release(transfer_state, s3_buffer);
        }
    }
    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        buffer_filled = -1;
    }

done:
    GlobusS3ExitInt(buffer_filled);
    return buffer_filled;
}
/* globus_l_s3_recv_single_data() */

static
void
globus_l_s3_recv_single_read_callback(
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t                      *buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void                               *user_arg)
{
    globus_i_s3_transfer_t             *transfer_state = user_arg;
    globus_i_s3_buffer_t               *s3_buffer;
    globus_i_s3_buffer_t               *tmp;
    globus_i_s3_buffer_t              **prev_next;
    bool                                signal = eof;

    GlobusS3Enter();

    GlobusS3Debug("recv_read_callback: op=%p result=%#x buffer=%p nbytes=%zd offset=%"PRIuMAX" eof=%d user_arg=%p\n",
            (void *) op,
            result,
            (void *) buffer,
            (size_t) nbytes,
            (uintmax_t) offset,
            (int) eof,
            user_arg);

    globus_mutex_lock(&transfer_state->mutex);

    transfer_state->eof |= eof;

    /* Remove from the registered buffer list */
    s3_buffer = transfer_state->registered_buffers;
    prev_next = &transfer_state->registered_buffers;

    while (s3_buffer != NULL)
    {
        if (s3_buffer->buffer == buffer)
        {
            *prev_next = s3_buffer->next;
            break;
        }
        prev_next = &s3_buffer->next;
        s3_buffer = s3_buffer->next;
    }

    assert(s3_buffer != NULL);

    /* Update the info about this buffer */
    s3_buffer->transfer_offset = offset;
    s3_buffer->buffer_used = nbytes;

    globus_gridftp_server_update_bytes_recvd(
            transfer_state->op,
            nbytes);

    /*
     * If something bad occurred, we will discard this block and flag
     * the request with an error
     */
    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        transfer_state->eof = true;
        goto bad_read;
    }

    if ((uint64_t) (offset + nbytes) > 
        (uint64_t) transfer_state->transfer_info->alloc_size)
    {
        /* peer sent me more than I told I'd send to s3  */
        transfer_state->result = GlobusGFSErrorSystemError("write", EFBIG);
        transfer_state->eof = true;
        goto bad_read;
    }

    if (s3_buffer->buffer_used == 0)
    {
        /* empty buffer */
        s3_buffer->transfer_offset = UINT64_C(-1);

        s3_buffer->next = transfer_state->free_buffers;
        transfer_state->free_buffers = s3_buffer;
    }
    else
    {
        /* Add it to the pending buffer list (in order) */
        tmp = transfer_state->pending_buffers;
        prev_next = &transfer_state->pending_buffers;
        while (tmp != NULL)
        {
            if ((tmp->transfer_offset + tmp->buffer_used)
                        <= s3_buffer->transfer_offset)
            {
                prev_next = &tmp->next;
                tmp = tmp->next;
            }
            else
            {
                break;
            }
        }
        /* tmp is the first buffer in pending that starts after s3_buffer 
         * (may be null) */
        s3_buffer->next = tmp;
        if (tmp == NULL)
        {
            transfer_state->pending_buffers_last = &s3_buffer->next;
        }

        *prev_next = s3_buffer;

        signal = true;
    }

    if (signal)
    {
        globus_cond_signal(&transfer_state->cond);
    }
    globus_mutex_unlock(&transfer_state->mutex);

    GlobusS3Exit();
    return;

bad_read:
    /* Since this was a bad read, we'll ignore the data */
    s3_buffer->buffer_used = 0;
    s3_buffer->transfer_offset = 0;
    s3_buffer->next = transfer_state->free_buffers;
    transfer_state->free_buffers = s3_buffer;

    globus_cond_signal(&transfer_state->cond);
    globus_mutex_unlock(&transfer_state->mutex);

    GlobusS3Exit();
    return;
}
/* globus_l_s3_recv_single_read_callback() */
// vim: ft=c:
